import os
from pathlib import Path

from django.test.html import parse_html

from crispy_forms.utils import render_crispy_form

TEST_DIR = os.path.dirname(os.path.abspath(__file__))


def parse_expected(expected_file):
    test_file = Path(TEST_DIR) / "results" / expected_file
    with test_file.open() as f:
        return parse_html(f.read())


def parse_form(form):
    html = render_crispy_form(form)
    return parse_html(html)



import pytest

from crispy_forms.layout import HTML, Div, Field, Fieldset, Layout, Submit


@pytest.fixture
def advanced_layout():
    return Layout(
        Div(
            Div(Div("email")),
            Div(Field("password1")),
            Submit("save", "save"),
            Fieldset(
                "legend",
                "first_name",
                HTML("extra text"),
            ),
            Layout("password2"),
        ),
        "last_name",
    )



import pytest
from django import forms
from django.forms.models import formset_factory, modelformset_factory
from django.shortcuts import render
from django.template import Context, Template
from django.test.html import parse_html
from django.test.utils import override_settings
from django.utils.translation import gettext_lazy as _

from crispy_forms.bootstrap import Field, InlineCheckboxes
from crispy_forms.helper import FormHelper
from crispy_forms.layout import HTML, Column, Fieldset, Layout, Row
from crispy_forms.utils import render_crispy_form

from .forms import (
    CheckboxesSampleForm,
    CrispyEmptyChoiceTestModel,
    CrispyTestModel,
    SampleForm,
    SampleForm3,
    SampleForm4,
    SampleForm6,
)
from .test_settings import TEMPLATE_DIRS
from .utils import parse_expected, parse_form


def test_invalid_unicode_characters(settings):
    # Adds a BooleanField that uses non valid unicode characters "ñ"
    form_helper = FormHelper()
    form_helper.add_layout(Layout("españa"))

    template = Template("""
        {% load crispy_forms_tags %}
        {% crispy form form_helper %}
    """)
    c = Context({"form": SampleForm(), "form_helper": form_helper})
    settings.CRISPY_FAIL_SILENTLY = False
    with pytest.raises(Exception):
        template.render(c)


def test_unicode_form_field():
    class UnicodeForm(forms.Form):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.fields["contraseña"] = forms.CharField()

        helper = FormHelper()
        helper.layout = Layout("contraseña")

    html = render_crispy_form(UnicodeForm())
    assert 'id="id_contraseña"' in html


def test_meta_extra_fields_with_missing_fields():
    class FormWithMeta(SampleForm):
        class Meta:
            fields = ("email", "first_name", "last_name")

    form = FormWithMeta()
    # We remove email field on the go
    del form.fields["email"]

    form_helper = FormHelper()
    form_helper.layout = Layout("first_name")

    template = Template("""
        {% load crispy_forms_tags %}
        {% crispy form form_helper %}
    """)
    c = Context({"form": form, "form_helper": form_helper})
    html = template.render(c)
    assert "email" not in html


def test_layout_unresolved_field(settings):
    form_helper = FormHelper()
    form_helper.add_layout(Layout("typo"))

    template = Template("""
        {% load crispy_forms_tags %}
        {% crispy form form_helper %}
    """)
    c = Context({"form": SampleForm(), "form_helper": form_helper})
    settings.CRISPY_FAIL_SILENTLY = False
    with pytest.raises(Exception):
        template.render(c)


def test_double_rendered_field(settings):
    form_helper = FormHelper()
    form_helper.add_layout(Layout("is_company", "is_company"))

    template = Template("""
        {% load crispy_forms_tags %}
        {% crispy form form_helper %}
    """)
    c = Context({"form": SampleForm(), "form_helper": form_helper})
    settings.CRISPY_FAIL_SILENTLY = False
    with pytest.raises(Exception):
        template.render(c)


def test_context_pollution():
    class ExampleForm(forms.Form):
        comment = forms.CharField()

    form = ExampleForm()
    form2 = SampleForm()

    template = Template("""
        {% load crispy_forms_tags %}
        {{ form.as_ul }}
        {% crispy form2 %}
        {{ form.as_ul }}
    """)
    c = Context({"form": form, "form2": form2})
    html = template.render(c)

    assert html.count('name="comment"') == 2
    assert html.count('name="is_company"') == 1


def test_layout_fieldset_row_html_with_unicode_fieldnames():
    form_helper = FormHelper()
    form_helper.add_layout(
        Layout(
            Fieldset(
                "Company Data",
                "is_company",
                css_id="fieldset_company_data",
                css_class="fieldsets",
                title="fieldset_title",
                test_fieldset="123",
            ),
            Fieldset(
                "User Data",
                "email",
                Row(
                    "password1",
                    "password2",
                    css_id="row_passwords",
                    css_class="rows",
                ),
                HTML('<a href="#" id="testLink">test link</a>'),
                HTML("""
                    {% if flag %}{{ message }}{% endif %}
                """),
                "first_name",
                "last_name",
            ),
        )
    )

    template = Template("""
        {% load crispy_forms_tags %}
        {% crispy form form_helper %}
    """)
    c = Context({"form": SampleForm(), "form_helper": form_helper, "flag": True, "message": "Hello!"})
    html = template.render(c)

    assert 'id="fieldset_company_data"' in html
    assert 'class="fieldsets' in html
    assert 'title="fieldset_title"' in html
    assert 'test-fieldset="123"' in html
    assert 'id="row_passwords"' in html
    assert html.count("<label") == 6
    assert 'class="row rows"' in html
    assert "Hello!" in html
    assert "testLink" in html


def test_change_layout_dynamically_delete_field():
    template = Template("""
        {% load crispy_forms_tags %}
        {% crispy form form_helper %}
    """)

    form = SampleForm()
    form.helper.add_layout(
        Layout(
            Fieldset(
                "Company Data",
                "is_company",
                "email",
                "password1",
                "password2",
                css_id="multifield_info",
            ),
            Column(
                "first_name",
                "last_name",
                css_id="column_name",
            ),
        )
    )

    # We remove email field on the go
    # Layout needs to be adapted for the new form fields
    del form.fields["email"]
    del form.helper.layout.fields[0].fields[1]

    c = Context({"form": form, "form_helper": form.helper})
    html = template.render(c)
    assert "email" not in html


def test_formset_layout():
    SampleFormSet = formset_factory(SampleForm, extra=3)
    formset = SampleFormSet()
    helper = FormHelper()
    helper.form_id = "thisFormsetRocks"
    helper.form_class = "formsets-that-rock"
    helper.form_method = "POST"
    helper.form_action = "simpleAction"
    helper.layout = Layout(
        Fieldset(
            "Item {{ forloop.counter }}",
            "is_company",
            "email",
        ),
        HTML("{% if forloop.first %}Note for first form only{% endif %}"),
        Row("password1", "password2"),
        Fieldset("", "first_name", "last_name"),
    )

    html = render_crispy_form(form=formset, helper=helper, context={"csrf_token": "aTestToken"})
    result = "test_formset_layout.html"
    assert parse_expected(result) == parse_html(html)


def test_modelformset_layout():
    CrispyModelFormSet = modelformset_factory(CrispyTestModel, form=SampleForm4, extra=3)
    formset = CrispyModelFormSet(queryset=CrispyTestModel.objects.none())
    formset.helper = FormHelper()
    formset.helper.layout = Layout("email")
    assert parse_expected("test_modelformset_layout.html") == parse_form(formset)


def test_i18n():
    template = Template("""
        {% load crispy_forms_tags %}
        {% crispy form form.helper %}
    """)
    form = SampleForm()
    form.helper.layout = Layout(
        HTML(_("i18n text")),
        Fieldset(
            _("i18n legend"),
            "first_name",
            "last_name",
        ),
    )
    html = template.render(Context({"form": form}))
    assert html.count("i18n legend") == 1


def test_default_layout():
    test_form = SampleForm()
    assert test_form.helper.layout.fields == [
        "is_company",
        "email",
        "password1",
        "password2",
        "first_name",
        "last_name",
        "datetime_field",
    ]


def test_default_layout_two():
    test_form = SampleForm3()
    assert test_form.helper.layout.fields == ["email"]


def test_modelform_layout_without_meta():
    test_form = SampleForm4()
    test_form.helper.layout = Layout("email")
    html = render_crispy_form(test_form)

    assert "email" in html
    assert "password" not in html


def test_specialspaceless_not_screwing_intended_spaces():
    # see issue #250
    test_form = SampleForm()
    test_form.fields["email"].widget = forms.Textarea()
    test_form.helper.layout = Layout("email", HTML("<span>first span</span> <span>second span</span>"))
    html = render_crispy_form(test_form)
    assert "<span>first span</span> <span>second span</span>" in html


def test_choice_with_none_is_selected():
    # see issue #701
    model_instance = CrispyEmptyChoiceTestModel()
    model_instance.fruit = None
    test_form = SampleForm6(instance=model_instance)
    html = render_crispy_form(test_form)
    assert "checked" in html


@override_settings(
    TEMPLATES=[
        {
            "BACKEND": "django.template.backends.django.DjangoTemplates",
            "DIRS": TEMPLATE_DIRS,
            "OPTIONS": {
                "loaders": [
                    "django.template.loaders.filesystem.Loader",
                    "django.template.loaders.app_directories.Loader",
                ],
            },
        }
    ]
)
def test_keepcontext_context_manager():
    # Test case for issue #180
    # Apparently it only manifest when using render_to_response this exact way
    form = CheckboxesSampleForm()
    # We use here InlineCheckboxes as it updates context in an unsafe way
    form.helper.layout = Layout("checkboxes", InlineCheckboxes("alphacheckboxes"), "numeric_multiple_checkboxes")
    context = {"form": form}

    response = render(request=None, template_name="crispy_render_template.html", context=context)

    assert response.content.count(b"checkbox-inline") == 3


def test_update_attributes_class():
    form = SampleForm()
    form.helper.layout = Layout("email", Field("password1"), "password2")
    form.helper["password1"].update_attributes(css_class="hello")
    html = render_crispy_form(form)
    assert html.count(' class="hello') == 1
    form.helper.layout = Layout(
        "email",
        Field("password1", css_class="hello"),
        "password2",
    )
    form.helper["password1"].update_attributes(css_class="hello2")
    html = render_crispy_form(form)
    assert html.count(' class="hello hello2') == 1



from django import forms
from django.template import Context, Template
from django.utils.translation import activate, deactivate
from django.utils.translation import gettext as _

from crispy_forms.bootstrap import (
    Accordion,
    AccordionGroup,
    Alert,
    AppendedText,
    Container,
    InlineCheckboxes,
    InlineRadios,
    PrependedAppendedText,
    PrependedText,
    Tab,
    TabHolder,
)
from crispy_forms.helper import FormHelper
from crispy_forms.layout import HTML, Field, Fieldset, Layout, MultiWidgetField
from crispy_forms.utils import render_crispy_form

from .forms import (
    CheckboxesSampleForm,
    CustomCheckboxSelectMultiple,
    CustomRadioSelect,
    SampleForm,
    SampleFormCustomWidgets,
)


def test_field_with_custom_template():
    test_form = SampleForm()
    test_form.helper.layout = Layout(Field("email", template="custom_field_template.html"))

    html = render_crispy_form(test_form)
    assert "<h1>Special custom field</h1>" in html


def test_multiwidget_field():
    template = Template("""
        {% load crispy_forms_tags %}
        {% crispy form %}
    """)

    test_form = SampleForm()
    test_form.helper.layout = Layout(
        MultiWidgetField(
            "datetime_field",
            attrs=({"rel": "test_dateinput"}, {"rel": "test_timeinput", "style": "width: 30px;", "type": "hidden"}),
        )
    )

    c = Context({"form": test_form})

    html = template.render(c)

    assert html.count('class="dateinput') == 1
    assert html.count('rel="test_dateinput"') == 1
    assert html.count('rel="test_timeinput"') == 2
    assert html.count('style="width: 30px;"') == 2
    assert html.count('type="hidden"') == 2


def test_field_type_hidden():
    template = Template("""
        {% load crispy_forms_tags %}
        {% crispy test_form %}
    """)

    test_form = SampleForm()
    test_form.helper.layout = Layout(
        Field("email", type="hidden", data_test=12),
        Field("datetime_field"),
    )

    c = Context({"test_form": test_form})
    html = template.render(c)

    # Check form parameters
    assert html.count('data-test="12"') == 1
    assert html.count('name="email"') == 1
    assert html.count('class="dateinput') == 1
    assert html.count('class="timeinput') == 1


def test_field_wrapper_class():
    form = SampleForm()
    form.helper.layout = Layout(Field("email", wrapper_class="testing"))

    html = render_crispy_form(form)
    assert html.count('class="form-group testing"') == 1


def test_html_with_carriage_returns():
    test_form = SampleForm()
    test_form.helper.layout = Layout(HTML("""
            if (a==b){
                // some comment
                a+1;
                foo();
            }
        """))
    html = render_crispy_form(test_form)
    assert html.count("\n") == 27


def test_i18n():
    activate("es")
    form = SampleForm()
    form.helper.layout = Layout(HTML(_("Enter a valid value.")))
    html = render_crispy_form(form)
    assert "Introduzca un valor válido" in html

    deactivate()


def test_remove_labels():
    form = SampleForm()
    # remove boolean field as label is still printed in bootstrap
    del form.fields["is_company"]

    for fields in form:
        fields.label = False

    html = render_crispy_form(form)

    assert "<label" not in html


def test_passthrough_context():
    """
    Test to ensure that context is passed through implicitly from outside of
    the crispy form into the crispy form templates.
    """
    form = SampleForm()
    form.helper.layout = Layout(
        Fieldset("", template="custom_fieldset_template_with_context.html"),
    )

    c = {"fieldset_context_var": "fieldset_context_value"}

    html = render_crispy_form(form, helper=form.helper, context=c)
    assert "Got context var: fieldset_context_value" in html


class TestBootstrapLayoutObjects:
    def test_custom_django_widget(self):
        # Make sure an inherited RadioSelect gets rendered as it
        form = SampleFormCustomWidgets()
        assert isinstance(form.fields["inline_radios"].widget, CustomRadioSelect)
        form.helper.layout = Layout("inline_radios")

        html = render_crispy_form(form)
        assert 'class="radio"' in html

        # Make sure an inherited CheckboxSelectMultiple gets rendered as it
        assert isinstance(form.fields["checkboxes"].widget, CustomCheckboxSelectMultiple)
        form.helper.layout = Layout("checkboxes")
        html = render_crispy_form(form)
        assert 'class="checkbox"' in html

    def test_inline_radios(self):
        test_form = CheckboxesSampleForm()
        test_form.helper.layout = Layout(InlineRadios("inline_radios"))
        html = render_crispy_form(test_form)
        assert html.count('radio-inline"') == 2

    def test_accordion_and_accordiongroup(self):
        test_form = SampleForm()
        test_form.helper.layout = Layout(
            Accordion(
                AccordionGroup("one", "first_name"),
                AccordionGroup("two", "password1", "password2"),
            )
        )
        html = render_crispy_form(test_form)

        assert html.count('<div class="panel panel-default"') == 2
        assert html.count('<div class="panel-group"') == 1
        assert html.count('<div class="panel-heading">') == 2
        assert html.count('<div id="one"') == 1
        assert html.count('<div id="two"') == 1
        assert html.count('name="first_name"') == 1
        assert html.count('name="password1"') == 1
        assert html.count('name="password2"') == 1

    def test_accordion_active_false_not_rendered(self):
        test_form = SampleForm()
        test_form.helper.layout = Layout(
            Accordion(
                AccordionGroup("one", "first_name"),
                # there is no ``active`` kwarg here.
            )
        )

        # The first time, there should be one of them there.
        html = render_crispy_form(test_form)

        accordion_class = "panel-collapse collapse in"

        assert html.count('<div id="one" class="%s"' % accordion_class) == 1

        test_form.helper.layout = Layout(
            Accordion(
                AccordionGroup("one", "first_name", active=False),
            )  # now ``active`` manually set as False
        )

        # This time, it shouldn't be there at all.
        html = render_crispy_form(test_form)
        assert html.count('<div id="one" class="%s collapse in"' % accordion_class) == 0

    def test_alert(self):
        test_form = SampleForm()
        test_form.helper.layout = Layout(Alert(content="Testing..."))
        html = render_crispy_form(test_form)

        assert html.count('<div class="alert"') == 1
        assert html.count('<button type="button" class="close"') == 1
        assert html.count("Testing...") == 1

    def test_alert_block(self):
        test_form = SampleForm()
        test_form.helper.layout = Layout(Alert(content="Testing...", block=True))
        html = render_crispy_form(test_form)

        assert html.count('<div class="alert alert-block"') == 1
        assert html.count("Testing...") == 1

    def test_tab_and_tab_holder(self):
        test_form = SampleForm()
        test_form.helper.layout = Layout(
            TabHolder(
                Tab("one", "first_name", css_id="custom-name", css_class="first-tab-class active"),
                Tab("two", "password1", "password2"),
            )
        )
        html = render_crispy_form(test_form)

        assert (
            html.count(
                '<ul class="nav nav-tabs"> <li class="tab-pane active">'
                '<a href="#custom-name" data-toggle="tab">One</a></li>'
            )
            == 1
        )
        assert html.count('<li class="tab-pane') == 2
        assert html.count("tab-pane") == 4

        assert html.count('class="tab-pane first-tab-class active"') == 1

        assert html.count('<div id="custom-name"') == 1
        assert html.count('<div id="two"') == 1
        assert html.count('name="first_name"') == 1
        assert html.count('name="password1"') == 1
        assert html.count('name="password2"') == 1

    def test_tab_helper_reuse(self):
        # this is a proper form, according to the docs.
        # note that the helper is a class property here,
        # shared between all instances
        class SampleForm(forms.Form):
            val1 = forms.CharField(required=False)
            val2 = forms.CharField(required=True)
            helper = FormHelper()
            helper.layout = Layout(
                TabHolder(
                    Tab("one", "val1"),
                    Tab("two", "val2"),
                )
            )

        # first render of form => everything is fine
        test_form = SampleForm()
        html = render_crispy_form(test_form)

        # second render of form => first tab should be active,
        # but not duplicate class
        test_form = SampleForm()
        html = render_crispy_form(test_form)
        assert html.count('class="nav-item active active"') == 0

        # render a new form, now with errors
        test_form = SampleForm(data={"val1": "foo"})
        html = render_crispy_form(test_form)
        tab_class = "tab-pane"
        # if settings.CRISPY_TEMPLATE_PACK == 'bootstrap4':
        # tab_class = 'nav-link'
        # else:
        # tab_class = 'tab-pane'
        # tab 1 should not be active
        assert html.count(f'<div id="one" \n    class="{tab_class} active') == 0
        # tab 2 should be active
        assert html.count(f'<div id="two" \n    class="{tab_class} active') == 1

    def test_radio_attrs(self):
        form = CheckboxesSampleForm()
        form.fields["inline_radios"].widget.attrs = {"class": "first"}
        form.fields["checkboxes"].widget.attrs = {"class": "second"}
        html = render_crispy_form(form)
        assert 'class="first"' in html
        assert 'class="second"' in html

    def test_hidden_fields(self):
        form = SampleForm()
        # All fields hidden
        for field in form.fields:
            form.fields[field].widget = forms.HiddenInput()

        form.helper.layout = Layout(
            AppendedText("password1", "foo"),
            PrependedText("password2", "bar"),
            PrependedAppendedText("email", "bar"),
            InlineCheckboxes("first_name"),
            InlineRadios("last_name"),
        )
        html = render_crispy_form(form)
        assert html.count("<input") == 5
        assert html.count('type="hidden"') == 5
        assert html.count("<label") == 0

    def test_multiplecheckboxes(self):
        test_form = CheckboxesSampleForm()
        html = render_crispy_form(test_form)

        assert html.count("checked") == 6

        test_form.helper[1].wrap(InlineCheckboxes, inline=True)
        html = render_crispy_form(test_form)

        assert html.count('inline="True"') == 4
        assert html.count('checkbox-inline"') == 3

    def test_multiple_checkboxes_unique_ids(self):
        test_form = CheckboxesSampleForm()
        html = render_crispy_form(test_form)

        expected_ids = [
            "checkboxes_0",
            "checkboxes_1",
            "checkboxes_2",
            "alphacheckboxes_0",
            "alphacheckboxes_1",
            "alphacheckboxes_2",
            "numeric_multiple_checkboxes_0",
            "numeric_multiple_checkboxes_1",
            "numeric_multiple_checkboxes_2",
        ]
        for id_suffix in expected_ids:
            expected_str = f'id="id_{id_suffix}"'
            assert html.count(expected_str) == 1

    def test_non_ascii_chars_in_container_name(self):
        """
        Test if non-ASCII characters are saved as css_id property.
        """
        name = "テスト"
        test_container = Container(name, "val1", "val2")
        assert test_container.css_id == name



import pytest
from django import forms
from django.conf import settings
from django.template.base import Template
from django.template.context import Context
from django.test import override_settings

from crispy_forms.helper import FormHelper
from crispy_forms.layout import Layout
from crispy_forms.templatetags.crispy_forms_filters import optgroups
from crispy_forms.utils import get_template_pack, list_difference, list_intersection, render_field

from .forms import GroupedChoiceForm, SampleForm, SampleForm5
from .utils import parse_expected, parse_form


def test_list_intersection():
    assert list_intersection([1, 3], [2, 3]) == [3]


def test_list_difference():
    assert list_difference([3, 1, 2, 3], [4, 1]) == [3, 2]


def test_render_field_with_none_field():
    rendered = render_field(field=None, form=None, context=None)
    assert rendered == ""


def test_custom_bound_field():
    from django.forms.boundfield import BoundField

    extra = "xyxyxyxyxyx"

    class CustomBoundField(BoundField):
        @property
        def auto_id(self):
            return extra

    class MyCharField(forms.CharField):
        def get_bound_field(self, form, field_name):
            return CustomBoundField(form, self, field_name)

    class MyForm(forms.Form):
        f = MyCharField()

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.helper = FormHelper()
            self.helper.layout = Layout("f")

    template = Template('{% load crispy_forms_tags %}\n{% crispy form "bootstrap3" %}')
    rendered = template.render(Context({"form": MyForm(data={"f": "something"})}))

    assert extra in rendered


def test_parse_expected_and_form():
    form = SampleForm()
    form.helper.layout = Layout("is_company")
    assert parse_form(form) == parse_expected("utils_test.html")


def test_optgroup_filter_nested():
    form = GroupedChoiceForm({"checkbox_select_multiple": ["cd", "vhs"]})
    form.as_p()
    groups = optgroups(form["checkbox_select_multiple"])
    audio, video, unknown = groups
    label, options, index = audio
    assert label == "Audio"
    assert options == [
        {
            "value": "vinyl",
            "type": "checkbox",
            "attrs": {"id": "id_checkbox_select_multiple_0_0"},
            "index": "0_0",
            "label": "Vinyl",
            "template_name": "django/forms/widgets/checkbox_option.html",
            "name": "checkbox_select_multiple",
            "selected": False,
            "wrap_label": True,
        },
        {
            "value": "cd",
            "type": "checkbox",
            "attrs": {"checked": True, "id": "id_checkbox_select_multiple_0_1"},
            "index": "0_1",
            "label": "CD",
            "template_name": "django/forms/widgets/checkbox_option.html",
            "name": "checkbox_select_multiple",
            "selected": True,
            "wrap_label": True,
        },
    ]
    assert index == 0
    label, options, index = video
    assert label == "Video"
    assert options == [
        {
            "value": "vhs",
            "template_name": "django/forms/widgets/checkbox_option.html",
            "label": "VHS Tape",
            "attrs": {"checked": True, "id": "id_checkbox_select_multiple_1_0"},
            "index": "1_0",
            "name": "checkbox_select_multiple",
            "selected": True,
            "type": "checkbox",
            "wrap_label": True,
        },
        {
            "value": "dvd",
            "template_name": "django/forms/widgets/checkbox_option.html",
            "label": "DVD",
            "attrs": {"id": "id_checkbox_select_multiple_1_1"},
            "index": "1_1",
            "name": "checkbox_select_multiple",
            "selected": False,
            "type": "checkbox",
            "wrap_label": True,
        },
    ]
    assert index == 1
    label, options, index = unknown
    assert label is None
    assert options == [
        {
            "value": "unknown",
            "selected": False,
            "template_name": "django/forms/widgets/checkbox_option.html",
            "label": "Unknown",
            "attrs": {"id": "id_checkbox_select_multiple_2"},
            "index": "2",
            "name": "checkbox_select_multiple",
            "type": "checkbox",
            "wrap_label": True,
        }
    ]
    assert index == 2


def test_optgroup_filter():
    form = SampleForm5({"checkbox_select_multiple": "1"})
    groups = optgroups(form["checkbox_select_multiple"])
    group = groups[0]
    label, option, index = group
    attrs = {"aria-invalid": "true", "checked": True, "id": "id_checkbox_select_multiple_0"}
    assert label is None
    assert option == [
        {
            "name": "checkbox_select_multiple",
            "value": 1,
            "label": 1,
            "selected": True,
            "index": "0",
            "attrs": attrs,
            "type": "checkbox",
            "template_name": "django/forms/widgets/checkbox_option.html",
            "wrap_label": True,
        }
    ]
    assert index == 0

    form = SampleForm5({"checkbox_select_multiple": 1})
    groups = optgroups(form["checkbox_select_multiple"])
    group = groups[0]
    label, option, index = group
    assert label is None
    assert option == [
        {
            "name": "checkbox_select_multiple",
            "value": 1,
            "label": 1,
            "selected": True,
            "index": "0",
            "attrs": attrs,
            "type": "checkbox",
            "template_name": "django/forms/widgets/checkbox_option.html",
            "wrap_label": True,
        }
    ]
    assert index == 0

    form = SampleForm5({"checkbox_select_multiple": ""})
    groups = optgroups(form["checkbox_select_multiple"])
    group = groups[0]
    label, option, index = group
    assert label is None
    attrs = {"id": "id_checkbox_select_multiple_0", "aria-invalid": "true"}
    assert option == [
        {
            "name": "checkbox_select_multiple",
            "value": 1,
            "label": 1,
            "selected": False,
            "index": "0",
            "attrs": attrs,
            "type": "checkbox",
            "template_name": "django/forms/widgets/checkbox_option.html",
            "wrap_label": True,
        }
    ]
    assert index == 0

    form = SampleForm5({"checkbox_select_multiple": None})
    groups = optgroups(form["checkbox_select_multiple"])
    group = groups[0]
    label, option, index = group
    assert label is None
    assert option == [
        {
            "name": "checkbox_select_multiple",
            "value": 1,
            "label": 1,
            "selected": False,
            "index": "0",
            "attrs": attrs,
            "type": "checkbox",
            "template_name": "django/forms/widgets/checkbox_option.html",
            "wrap_label": True,
        }
    ]
    assert index == 0


@override_settings()
def test_get_template_pack():
    del settings.CRISPY_TEMPLATE_PACK
    with pytest.raises(AttributeError):
        get_template_pack()
    with pytest.raises(AttributeError):
        settings.CRISPY_TEMPLATE_PACK



import pytest
from django import forms

from crispy_forms.bootstrap import AppendedText
from crispy_forms.exceptions import DynamicError
from crispy_forms.helper import FormHelper, FormHelpersException
from crispy_forms.layout import HTML, Div, Field, Fieldset, Layout, MultiField, Pointer

from .forms import SampleForm


def test_wrap_all_fields():
    helper = FormHelper()
    layout = Layout("email", "password1", "password2")
    helper.layout = layout

    helper.all().wrap(Field, css_class="test-class")
    for field in layout.fields:
        assert isinstance(field, Field)
        assert field.attrs["class"] == "test-class"

    assert layout[0][0] == "email"
    assert layout[1][0] == "password1"
    assert layout[2][0] == "password2"


def test_wrap_selected_fields():
    helper = FormHelper()
    layout = Layout("email", "password1", "password2")
    helper.layout = layout

    helper[1:3].wrap(Field, css_class="test-class")
    assert not isinstance(layout.fields[0], Field)
    assert isinstance(layout.fields[1], Field)
    assert isinstance(layout.fields[2], Field)

    helper[0].wrap(Fieldset, "legend", css_class="test-class")
    assert isinstance(layout[0], Fieldset)
    assert layout[0].legend == "legend"
    assert layout[0][0] == "email"


def test_wrap_together_with_slices():
    helper = FormHelper()
    layout = Layout("email", "password1", "password2")
    helper.layout = layout
    helper[1:3].wrap_together(Field, css_class="test-class")
    assert layout.fields[0] == "email"
    assert isinstance(layout.fields[1], Field)
    assert layout.fields[1][0] == "password1"
    assert layout.fields[1][1] == "password2"

    layout = Layout(
        Div("email"),
        "password1",
        "password2",
    )
    helper.layout = layout
    helper[0:3].wrap_together(Field, css_class="test-class")
    assert isinstance(layout.fields[0], Field)
    assert isinstance(layout.fields[0][0], Div)
    assert layout.fields[0][0][0] == "email"
    assert layout.fields[0][1] == "password1"
    assert layout.fields[0][2] == "password2"

    layout = Layout("email", "password1", "password2")
    helper.layout = layout
    helper[0].wrap_together(Field, css_class="test-class")
    assert isinstance(layout.fields[0], Field)
    assert layout.fields[1] == "password1"
    assert layout.fields[2] == "password2"

    layout = Layout("email", "password1", "password2")
    helper.layout = layout
    helper[0].wrap_together(Fieldset, "legend", css_class="test-class")
    assert isinstance(layout.fields[0], Fieldset)
    assert layout.fields[0].legend == "legend"
    assert layout.fields[1] == "password1"
    assert layout.fields[2] == "password2"


def test_wrap_together_partial_slices():
    helper = FormHelper()
    layout = Layout("email", "password1", "password2")
    helper.layout = layout

    helper[:2].wrap_together(Field, css_class="test-class")
    assert isinstance(layout.fields[0], Field)
    assert layout.fields[1] == "password2"
    assert layout.fields[0][0] == "email"
    assert layout.fields[0][1] == "password1"

    helper = FormHelper()
    layout = Layout("email", "password1", "password2")
    helper.layout = layout

    helper[1:].wrap_together(Field, css_class="test-class")
    assert layout.fields[0] == "email"
    assert isinstance(layout.fields[1], Field)
    assert layout.fields[1][0] == "password1"
    assert layout.fields[1][1] == "password2"


def test_update_attributes():
    helper = FormHelper()
    helper.layout = Layout("email", Field("password1"), "password2")
    helper["password1"].update_attributes(readonly=True)
    assert "readonly" in helper.layout[1].attrs


def test_update_attributes_and_wrap_once():
    helper = FormHelper()
    layout = Layout("email", Field("password1"), "password2")
    helper.layout = layout
    helper.filter(Field).update_attributes(readonly=True)
    assert isinstance(layout[1], Field)
    assert layout[1].attrs == {"readonly": True}

    layout = Layout(
        "email",
        Div(Field("password1")),
        "password2",
    )
    helper.layout = layout
    helper.filter(Field, max_level=2).update_attributes(readonly=True)
    assert isinstance(layout[1][0], Field)
    assert layout[1][0].attrs == {"readonly": True}

    layout = Layout(
        "email",
        Div(Field("password1")),
        "password2",
    )
    helper.layout = layout

    helper.filter(str, greedy=True).wrap_once(Field)
    helper.filter(Field, greedy=True).update_attributes(readonly=True)

    assert isinstance(layout[0], Field)
    assert isinstance(layout[1][0], Field)
    assert isinstance(layout[1][0][0], str)
    assert isinstance(layout[2], Field)
    assert layout[1][0].attrs == {"readonly": True}
    assert layout[0].attrs == {"readonly": True}
    assert layout[2].attrs == {"readonly": True}


def test_get_layout_objects():
    layout_1 = Layout(Div())
    assert layout_1.get_layout_objects(Div) == [Pointer([0], "div")]

    layout_2 = Layout(Div(Div(Div("email")), Div("password1"), "password2"))
    assert layout_2.get_layout_objects(Div) == [Pointer([0], "div")]
    assert layout_2.get_layout_objects(Div, max_level=1) == [
        Pointer([0], "div"),
        Pointer([0, 0], "div"),
        Pointer([0, 1], "div"),
    ]
    assert layout_2.get_layout_objects(Div, max_level=2) == [
        Pointer([0], "div"),
        Pointer([0, 0], "div"),
        Pointer([0, 0, 0], "div"),
        Pointer([0, 1], "div"),
    ]

    layout_3 = Layout(
        "email",
        Div("password1"),
        "password2",
    )
    assert layout_3.get_layout_objects(str, max_level=2) == [
        Pointer([0], "email"),
        Pointer([1, 0], "password1"),
        Pointer([2], "password2"),
    ]

    layout_4 = Layout(
        Div(
            Div("field_name"),
            "field_name2",
        ),
        Div("password"),
        "extra_field",
    )
    assert layout_4.get_layout_objects(Div) == [Pointer([0], "div"), Pointer([1], "div")]
    assert layout_4.get_layout_objects(Div, max_level=1) == [
        Pointer([0], "div"),
        Pointer([0, 0], "div"),
        Pointer([1], "div"),
    ]


def test_filter_and_wrap():
    helper = FormHelper()
    layout = Layout(
        "email",
        Div("password1"),
        "password2",
    )
    helper.layout = layout

    helper.filter(str).wrap(Field, css_class="test-class")
    assert isinstance(layout.fields[0], Field)
    assert isinstance(layout.fields[1], Div)
    assert isinstance(layout.fields[2], Field)
    assert layout[2][0] == "password2"

    # Wrapping a div in a div
    helper.filter(Div).wrap(Div, css_class="test-class")
    assert isinstance(layout.fields[1], Div)
    assert isinstance(layout.fields[1].fields[0], Div)
    assert layout[1][0][0] == "password1"


def test_filter_and_wrap_side_effects():
    helper = FormHelper()
    layout = Layout(
        Div(
            "extra_field",
            Div("password1"),
        ),
    )
    helper.layout = layout
    with pytest.raises(DynamicError):
        helper.filter(Div, max_level=2).wrap(Div, css_class="test-class")


def test_get_field_names():
    layout_1 = Div("field_name")
    assert layout_1.get_field_names() == [Pointer([0], "field_name")]

    layout_2 = Div(Div("field_name"))
    assert layout_2.get_field_names() == [Pointer([0, 0], "field_name")]

    layout_3 = Div(Div("field_name"), "password")
    assert layout_3.get_field_names() == [Pointer([0, 0], "field_name"), Pointer([1], "password")]

    layout_4 = Div(
        Div(
            Div("field_name"),
            "field_name2",
        ),
        Div("password"),
        "extra_field",
    )
    assert layout_4.get_field_names() == [
        Pointer([0, 0, 0], "field_name"),
        Pointer([0, 1], "field_name2"),
        Pointer([1, 0], "password"),
        Pointer([2], "extra_field"),
    ]

    layout_5 = Div(
        Div(
            "field_name",
            "field_name2",
        ),
        "extra_field",
    )
    assert layout_5.get_field_names() == [
        Pointer([0, 0], "field_name"),
        Pointer([0, 1], "field_name2"),
        Pointer([1], "extra_field"),
    ]


def test_layout_get_field_names():
    layout_1 = Layout(Div("field_name"), "password")
    assert layout_1.get_field_names() == [
        Pointer([0, 0], "field_name"),
        Pointer([1], "password"),
    ]

    layout_2 = Layout(Div("field_name"), "password", Fieldset("legend", "extra_field"))
    assert layout_2.get_field_names() == [
        Pointer([0, 0], "field_name"),
        Pointer([1], "password"),
        Pointer([2, 0], "extra_field"),
    ]

    layout_3 = Layout(Div(Div(Div("email")), Div("password1"), "password2"))
    assert layout_3.get_field_names() == [
        Pointer([0, 0, 0, 0], "email"),
        Pointer([0, 1, 0], "password1"),
        Pointer([0, 2], "password2"),
    ]


def test_filter_by_widget(advanced_layout):
    form = SampleForm()
    form.helper.layout = advanced_layout
    assert form.helper.filter_by_widget(forms.PasswordInput).slice == [
        Pointer([0, 1, 0, 0], "password1"),
        Pointer([0, 4, 0], "password2"),
    ]


def test_exclude_by_widget(advanced_layout):
    form = SampleForm()
    form.helper.layout = advanced_layout
    assert form.helper.exclude_by_widget(forms.PasswordInput).slice == [
        Pointer([0, 0, 0, 0], "email"),
        Pointer([0, 3, 0], "first_name"),
        Pointer([1], "last_name"),
    ]


def test_exclude_by_widget_and_wrap(advanced_layout):
    form = SampleForm()
    form.helper.layout = advanced_layout
    form.helper.exclude_by_widget(forms.PasswordInput).wrap(Field, css_class="hero")
    # Check wrapped fields
    assert isinstance(form.helper.layout[0][0][0][0], Field)
    assert isinstance(form.helper.layout[0][3][0], Field)
    assert isinstance(form.helper.layout[1], Field)
    # Check others stay the same
    assert isinstance(form.helper.layout[0][3][1], HTML)
    assert isinstance(form.helper.layout[0][1][0][0], str)
    assert isinstance(form.helper.layout[0][4][0], str)


def test_all_without_layout():
    form = SampleForm()
    form.helper = FormHelper()  # reset helper to remove default layout
    with pytest.raises(FormHelpersException):
        form.helper.all().wrap(Div)


def test_filter_by_widget_without_form(advanced_layout):
    form = SampleForm()
    form.helper = FormHelper()  # reset helper to remove default layout
    form.helper.layout = advanced_layout
    with pytest.raises(FormHelpersException):
        form.helper.filter_by_widget(forms.PasswordInput)


def test_formhelper__getitem__():
    helper = FormHelper()
    layout = Layout(
        Div("email"),
        "password1",
    )
    helper.layout = layout
    helper["email"].wrap(Field, css_class="hero")
    assert isinstance(layout[0][0], Field)
    assert layout[0][0][0] == "email"

    helper = FormHelper()
    helper.layout = Layout("password1")
    helper["password1"].wrap(AppendedText, "extra")
    assert isinstance(helper.layout[0], AppendedText)
    assert helper.layout[0][0] == "password1"
    assert helper.layout[0].text == "extra"


def test_formhelper__setitem__():
    helper = FormHelper()
    layout = Layout("first_field", Div("email"))
    helper.layout = layout
    helper[0] = "replaced"
    assert layout[0] == "replaced"


def test_formhelper__delitem__and__len__():
    helper = FormHelper()
    layout = Layout("first_field", Div("email"))
    helper.layout = layout
    del helper[0]
    assert len(helper) == 1


def test__delitem__and__len__layout_object():
    layout = Layout("first_field", Div("email"))
    del layout[0]
    assert len(layout) == 1


def test__getitem__layout_object():
    layout = Layout(Div(Div(Div("email")), Div("password1"), "password2"))
    assert isinstance(layout[0], Div)
    assert isinstance(layout[0][0], Div)
    assert isinstance(layout[0][0][0], Div)
    assert isinstance(layout[0][1], Div)
    assert isinstance(layout[0][1][0], str)
    assert isinstance(layout[0][2], str)


def test__getattr__append_layout_object():
    layout = Layout(Div("email"))
    layout.append("password1")
    assert isinstance(layout[0], Div)
    assert isinstance(layout[0][0], str)
    assert isinstance(layout[1], str)


def test__setitem__layout_object():
    layout = Layout(Div("email"))
    layout[0][0] = "password1"
    assert isinstance(layout[0], Div)
    assert layout[0][0] == "password1"


def test_filter():
    helper = FormHelper()
    helper.layout = Layout(
        Div(
            MultiField("field_name"),
            "field_name2",
        ),
        Div("password"),
        "extra_field",
    )
    assert helper.filter(Div, MultiField).slice == [Pointer([0], "div"), Pointer([1], "div")]
    assert helper.filter(Div, MultiField, max_level=1).slice == [
        Pointer([0], "div"),
        Pointer([0, 0], "multifield"),
        Pointer([1], "div"),
    ]
    assert helper.filter(MultiField, max_level=1).slice == [Pointer([0, 0], "multifield")]



import os

SITE_ROOT = os.path.join(os.path.dirname(os.path.abspath(__file__)))

TEMPLATE_LOADERS = (
    "django.template.loaders.filesystem.Loader",
    "django.template.loaders.app_directories.Loader",
)

TEMPLATE_DIRS = os.path.join(SITE_ROOT, "templates")

INSTALLED_APPS = "crispy_forms"
SECRET_KEY = "secretkey"
